import abc
import inspect
from typing import Any
from .error import ValidateError


# We use Ellipsis to differentiate if caller assign non-default value for kwargs
NO_CHECK = USE_DEFAULT = Ellipsis


class BelieveBase(abc.ABC):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.__arg_str = self.__init_arg_str(*args, **kwargs)

    def __ne__(self, rhs: Any) -> bool:
        return not self.__eq__(rhs)

    def __eq__(self, rhs: Any) -> bool:
        try:
            self.validate(rhs)
            return True
        except ValidateError:
            return False

    def __str__(self) -> str:
        return "%s(%s)" % (self.__class__.__name__, self.__arg_str)

    def __repr__(self) -> str:
        return self.__str__()

    def __to_str(self, val: Any) -> str:
        if inspect.isclass(val) and hasattr(val, "__name__"):
            return val.__name__
        else:
            return str(val)

    def __init_arg_str(self, *args: Any, **kwargs: Any) -> str:
        list_of_arg = list(args)
        for k, v in kwargs.items():
            if v != Ellipsis:  # only show overwritten kwargs
                list_of_arg.append("%s=%s" % (k, v))
        return ", ".join([self.__to_str(i) for i in list_of_arg])

    def raise_validate_error(self, rhs: Any, **kwargs: Any) -> None:
        raise ValidateError(self, rhs, **kwargs)

    @abc.abstractmethod
    def validate(self, rhs: Any, e_path: str = "") -> None:
        raise NotImplemented()


def validate(v1: Any, v2: Any, e_path: str = "") -> None:
    if isinstance(v1, BelieveBase):
        v1.validate(v2, e_path=e_path)
    elif isinstance(v2, BelieveBase):
        v2.validate(v1, e_path=e_path)
    elif v1 != v2:
        raise ValidateError(v1, v2, e_path=e_path)



from .internal import BelieveBase
from typing import Any


class Nullable(BelieveBase):
    def __init__(self, obj: Any) -> None:
        super().__init__(obj)
        self.__obj = obj

    def validate(self, rhs: Any, e_path: str = "") -> None:
        if rhs is None:
            return
        if self.__obj != rhs:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_param")


class Not(BelieveBase):
    def __init__(self, obj: Any) -> None:
        super().__init__(obj)
        self.__obj = obj

    def validate(self, rhs: Any, e_path: str = "") -> None:
        if self.__obj == rhs:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_param")


class Any(BelieveBase):
    def __init__(self, *type_classes: type) -> None:
        for type_class in type_classes:
            assert isinstance(type_class, type)

        super().__init__(*type_classes)
        self.__type_classes = type_classes

    def validate(self, rhs: Any, e_path: str = "") -> None:
        if not self.__type_classes:
            return
        if not isinstance(rhs, self.__type_classes):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_param")



import json
import unittest
import uuid
import os


import believe as B

class BaseClass:
    def gen_sha1(self):
        return os.urandom(20).hex()

    def fail_validate_with(self, v1, v2, xss_unsafe_message, xss_safe_message=""):
        try:
            B.validate(v1, v2)
            assert not "should fail"
        except B.ValidateError as e:
            assert e.xss_safe_message() == xss_safe_message
            assert e.xss_unsafe_message() == xss_unsafe_message
            assert str(e) == xss_unsafe_message


class TestNumberMatcher(BaseClass):
    def test_number_matcher__Almost_int(self):
        assert "Almost(1)" == str(B.Almost(1))
        assert 1 == B.Almost(0)
        assert 1 == B.Almost(2)
        assert 1 == B.Almost(3)
        assert 1 != B.Almost(5)
        assert None != B.Almost(0)

        self.fail_validate_with(10, B.Almost(1),
                                "[e_msg=not_in_range] 10 != Almost(1)",
                                "[e_msg=not_in_range]")
        self.fail_validate_with(10, B.Almost(1, ts_range=2),
                                "[e_msg=not_in_range] 10 != Almost(1, ts_range=2)",
                                "[e_msg=not_in_range]")

    def test_number_matcher__Almost_float(self):
        assert "Almost(1.1)" == str(B.Almost(1.1))
        assert 1.1 == B.Almost(0)
        assert 1.1 == B.Almost(2)
        assert 1.1 == B.Almost(3)
        assert 1.1 != B.Almost(5)

        self.fail_validate_with(10, B.Almost(1.1),
                                "[e_msg=not_in_range] 10 != Almost(1.1)",
                                "[e_msg=not_in_range]")
        self.fail_validate_with(10, B.Almost(1.1, ts_range=2.2),
                                "[e_msg=not_in_range] 10 != Almost(1.1, ts_range=2.2)",
                                "[e_msg=not_in_range]")

    def test_number_matcher__AnyInt(self):
        assert "AnyInt(min_value=1, max_value=2)" == str(B.AnyInt(min_value=1, max_value=2))
        assert 1 == B.AnyInt(min_value=1, max_value=2)
        assert 2 == B.AnyInt(min_value=1, max_value=2)
        assert 0 != B.AnyInt(min_value=1, max_value=2)
        assert 3 != B.AnyInt(min_value=1, max_value=2)
        assert None != B.AnyInt(min_value=1, max_value=2)
        assert "None" != B.AnyInt(min_value=1, max_value=2)

        self.fail_validate_with(3, B.AnyInt(min_value=1, max_value=2),
                                "[e_msg=value_too_large: 3 > 2] 3 != AnyInt(min_value=1, max_value=2)",
                                "[e_msg=value_too_large: 3 > 2]")
        self.fail_validate_with(0, B.AnyInt(min_value=1, max_value=2),
                                "[e_msg=value_too_small: 0 < 1] 0 != AnyInt(min_value=1, max_value=2)",
                                "[e_msg=value_too_small: 0 < 1]")

    def test_number_matcher__AnyFloat(self):
        assert "AnyFloat(min_value=1.0, max_value=2.0)" ==  str(B.AnyFloat(min_value=1.0, max_value=2.0))
        assert 1.0 == B.AnyFloat(min_value=1.0, max_value=2.0)
        assert 2.0 == B.AnyFloat(min_value=1.0, max_value=2.0)
        assert 0.0 != B.AnyFloat(min_value=1.0, max_value=2.0)
        assert 3.0 != B.AnyFloat(min_value=1.0, max_value=2.0)
        assert None != B.AnyFloat(min_value=1.0, max_value=2.0)
        assert "None" != B.AnyFloat(min_value=1.0, max_value=2.0)

        self.fail_validate_with(3.0, B.AnyFloat(min_value=1.0, max_value=2.0),
                                "[e_msg=value_too_large: 3.0 > 2.0] 3.0 != AnyFloat(min_value=1.0, max_value=2.0)",
                                "[e_msg=value_too_large: 3.0 > 2.0]")
        self.fail_validate_with(0.0, B.AnyFloat(min_value=1.0, max_value=2.0),
                                "[e_msg=value_too_small: 0.0 < 1.0] 0.0 != AnyFloat(min_value=1.0, max_value=2.0)",
                                "[e_msg=value_too_small: 0.0 < 1.0]")

class TestStrMatcher(BaseClass):

    def test_str_matcher__AnyIPV4(self):
        assert "AnyIPV4()" == str(B.AnyIPV4())
        assert '192.168.1.1' ==  B.AnyIPV4()
        assert '127.0.0' != B.AnyIPV4()
        assert '0.0.0.x' != B.AnyIPV4()
        assert "256.1.1.1" != B.AnyIPV4()
        assert 12345 != B.AnyIPV4()

        self.fail_validate_with(12345, B.AnyIPV4(),
                                "[e_msg=invalid_ipv4] 12345 != AnyIPV4()",
                                "[e_msg=invalid_ipv4]")

    def test_str_matcher__AnySHA1(self):
        assert "AnySHA1()" ==  str(B.AnySHA1())
        assert self.gen_sha1() == B.AnySHA1()
        assert '123' != B.AnySHA1()
        assert 123 != B.AnySHA1()
        assert "Z" + self.gen_sha1()[1:] != B.AnySHA1()

        self.fail_validate_with('123', B.AnySHA1(),
                                "[e_msg=invalid_sha1] '123' != AnySHA1()",
                                "[e_msg=invalid_sha1]")

    def test_str_matcher__AnyUUID(self):
        assert "AnyUUID()" == str(B.AnyUUID())
        assert str(uuid.uuid4()) == B.AnyUUID()
        assert None != B.AnyUUID()
        assert 10 != B.AnyUUID()
        assert "abc" != B.AnyUUID()

        self.fail_validate_with(10, B.AnyUUID(),
                                "[e_msg=invalid_uuid] 10 != AnyUUID()",
                                "[e_msg=invalid_uuid]")

    def test_str_matcher__AnyIntStr(self):
        assert "AnyIntStr()" == str(B.AnyIntStr())
        assert str(1) == B.AnyIntStr()
        assert None != B.AnyIntStr()
        assert 10 != B.AnyIntStr()
        assert 'invalid' !=  B.AnyIntStr()

        self.fail_validate_with(10, B.AnyIntStr(),
                                "[e_msg=not_string] 10 != AnyIntStr()",
                                "[e_msg=not_string]")
        self.fail_validate_with("NotInt", B.AnyIntStr(),
                                "[e_msg=not_int_string] 'NotInt' != AnyIntStr()",
                                "[e_msg=not_int_string]")

    def test_str_matcher__AnyStr(self):
        assert "AnyStr()" == str(B.AnyStr())
        assert "AnyStr(min_len=1, max_len=2)" == str(B.AnyStr(min_len=1, max_len=2))
        assert str(uuid.uuid4()) == B.AnyStr()
        assert None != B.AnyStr()
        assert 10 != B.AnyStr()
        assert "123" != B.AnyStr(min_len=4)
        assert "123" != B.AnyStr(max_len=2)

        self.fail_validate_with(10, B.AnyStr(), "[e_msg=not_string] 10 != AnyStr()", "[e_msg=not_string]")
        self.fail_validate_with("123", B.AnyStr(min_len=4),
                                "[e_msg=string_too_short: 3 < 4] '123' != AnyStr(min_len=4)",
                                "[e_msg=string_too_short: 3 < 4]")
        self.fail_validate_with("123", B.AnyStr(max_len=2),
                                "[e_msg=string_too_long: 3 > 2] '123' != AnyStr(max_len=2)",
                                "[e_msg=string_too_long: 3 > 2]")

    def test_str_matcher__AnyJsonStr(self):
        assert "AnyJsonStr({})" == str(B.AnyJsonStr({}))
        assert '{"foo": "bar", "bar": "foo"}' == B.AnyJsonStr({"foo": "bar", "bar": "foo"})

        self.fail_validate_with(b"{}", B.AnyJsonStr({}),
                               "[e_msg=not_string] b'{}' != AnyJsonStr({})",
                               "[e_msg=not_string]")
        self.fail_validate_with("{", B.AnyJsonStr({}),
                               "[e_msg=not_json_string] '{' != AnyJsonStr({})",
                               "[e_msg=not_json_string]")
        self.fail_validate_with("[]", B.AnyJsonStr({}),
                               "[e_msg=mismatch_json_string] '[]' != AnyJsonStr({})",
                               "[e_msg=mismatch_json_string]")

    def test_str_matcher__AnyJsonStr(self):
        assert "AnyJsonStr({})" == str(B.AnyJsonStr({}))
        assert '{"foo": "bar", "bar": "foo"}' == B.AnyJsonStr({"foo": "bar", "bar": "foo"})

        self.fail_validate_with(b"{}", B.AnyJsonStr({}),
                               "[e_msg=not_string] b'{}' != AnyJsonStr({})",
                               "[e_msg=not_string]")
        self.fail_validate_with("{", B.AnyJsonStr({}),
                               "[e_msg=not_json_string] '{' != AnyJsonStr({})",
                               "[e_msg=not_json_string]")
        self.fail_validate_with("[]", B.AnyJsonStr({}),
                               "[e_msg=mismatch_json_string] '[]' != AnyJsonStr({})",
                               "[e_msg=mismatch_json_string]")

    def test_str_matcher__AnyUrl(self):
        assert "AnyUrl(https://example.com)" == str(B.AnyUrl("https://example.com"))
        assert "https://example.com" == B.AnyUrl("https://example.com")

        # test default port
        assert "https://example.com:443" == B.AnyUrl("https://example.com")
        assert "http://example.com:80" == B.AnyUrl("http://example.com")

        # test query
        assert "https://example.com/path?q1=1&q2=2" == B.AnyUrl("https://example.com/path?q2=2&q1=1")

        self.fail_validate_with(b"https://example.com", B.AnyUrl("https://example.com"),
                               "[e_msg=not_string] b'https://example.com' != AnyUrl(https://example.com)",
                               "[e_msg=not_string]")

        self.fail_validate_with("http://example.com", B.AnyUrl("https://example.com"),
                               "[e_msg=mismatch_scheme] 'http://example.com' != AnyUrl(https://example.com)",
                               "[e_msg=mismatch_scheme]")

        self.fail_validate_with("https://user@example.com", B.AnyUrl("https://example.com"),
                               "[e_msg=mismatch_username] 'https://user@example.com' != AnyUrl(https://example.com)",
                               "[e_msg=mismatch_username]")

        self.fail_validate_with("https://user:pass@example.com", B.AnyUrl("https://user@example.com"),
                               "[e_msg=mismatch_password] 'https://user:pass@example.com' != AnyUrl(https://user@example.com)",
                               "[e_msg=mismatch_password]")

        self.fail_validate_with("https://foo.com", B.AnyUrl("https://example.com"),
                               "[e_msg=mismatch_hostname] 'https://foo.com' != AnyUrl(https://example.com)",
                               "[e_msg=mismatch_hostname]")

        self.fail_validate_with("https://example.com:8888", B.AnyUrl("https://example.com"),
                               "[e_msg=mismatch_port] 'https://example.com:8888' != AnyUrl(https://example.com)",
                               "[e_msg=mismatch_port]")

        self.fail_validate_with("https://example.com/abc", B.AnyUrl("https://example.com/def"),
                               "[e_msg=mismatch_path] 'https://example.com/abc' != AnyUrl(https://example.com/def)",
                               "[e_msg=mismatch_path]")

        self.fail_validate_with("https://example.com/foo?bar=1", B.AnyUrl("https://example.com/foo?bar=2"),
                               "[e_msg=mismatch_query] 'https://example.com/foo?bar=1' != AnyUrl(https://example.com/foo?bar=2)",
                               "[e_msg=mismatch_query]")


class TestListMatcher(BaseClass):

    def test_list_matcher__OneOf(self):
        assert "OneOf(1, 2, 3)" == str(B.OneOf(1, 2, 3))
        assert 1 == B.OneOf(1, 2, 3)
        assert 4 != B.OneOf(1, 2, 3)

        self.fail_validate_with(4, B.OneOf([2, 1, 3]),
                                "[e_msg=invalid_argument] 4 != OneOf([2, 1, 3])",
                                "[e_msg=invalid_argument]")

    def test_list_matcher__AnyOrder(self):
        assert "AnyOrder([2, 1, 1])" == str(B.AnyOrder([2, 1, 1]))
        assert [1, 1, 2] == B.AnyOrder([2, 1, 1])
        assert [1, 2, 3] == B.AnyOrder([2, 1, 3])
        assert [1, 2] != B.AnyOrder([2, 1, 3])
        assert [1, 2, 3] != B.AnyOrder([2, 1])

        self.fail_validate_with([1, 2], B.AnyOrder([2, 1, 3]),
                                "[e_msg=different_length] [1, 2] != AnyOrder([2, 1, 3])",
                                "[e_msg=different_length]")

    def test_list_matcher__ListOf(self):
        assert "ListOf(1)" == str(B.ListOf(1))
        assert [] == B.ListOf(1)
        assert [1, 1, 1] == B.ListOf(1)
        assert [1, 2, 3] != B.ListOf(1)
        assert [] == B.ListOf(1)
        assert () != B.ListOf(1)
        assert "aaa" != B.ListOf("a")
        assert ["a", "a"] == B.ListOf("a")

        self.fail_validate_with([1, 2, 3], B.ListOf(1),
                                "[e_path=$.1] 2 != 1",
                                "[e_path=$.1]")

        # dict
        assert [{"name": "123"}, {"name": "456"}] == B.ListOf({"name": B.Any(str)})
        assert [{"name": "123"}, {"name": "456", "type": "789"}] != B.ListOf({"name": B.Any(str)})

        self.fail_validate_with([{}], B.ListOf({"name": B.Any(str)}),
                                "[e_path=$.0] {} != {'name': Any(str)}",
                                "[e_path=$.0]")

        # n_item
        assert ["a", "a"] != B.ListOf("a", n_item=1)
        assert ["a", "a"] == B.ListOf("a", n_item=2)

        self.fail_validate_with([1], B.ListOf(1, n_item=2),
                                "[e_msg=mismatch_item_count: 1 != 2] [1] != ListOf(1, n_item=2)",
                                "[e_msg=mismatch_item_count: 1 != 2]")

        # min_item
        assert [] != B.ListOf("a", min_item=1)
        assert ["a"] == B.ListOf("a", min_item=1)
        assert ["a", "a"] == B.ListOf("a", min_item=1)

        self.fail_validate_with([1], B.ListOf(1, min_item=2),
                                "[e_msg=too_few_items: 1 < 2] [1] != ListOf(1, min_item=2)",
                                "[e_msg=too_few_items: 1 < 2]")

        # max_item
        assert [] == B.ListOf("a", max_item=1)
        assert ["a"] == B.ListOf("a", max_item=1)
        assert ["a", "a"] != B.ListOf("a", max_item=1)

        self.fail_validate_with([1], B.ListOf(1, max_item=0),
                                "[e_msg=too_many_items: 1 > 0] [1] != ListOf(1, max_item=0)",
                                "[e_msg=too_many_items: 1 > 0]")

class TestDictMatcher(BaseClass):

    def test_dict_matcher__Dict(self):
        exp_dict = B.Dict({"req1": 1,
                           "req2": 2,
                           "opt1": B.Optional(1),
                           "opt2": B.Optional(2)})

        assert "Dict({'req1': 1, 'req2': 2, 'opt1': Optional(1), 'opt2': Optional(2)})" == str(exp_dict)

        assert exp_dict == {"req1": 1, "req2": 2, "opt1": 1, "opt2": 2}
        assert exp_dict == {"req1": 1, "req2": 2, "opt1": 1}
        assert exp_dict == {"req1": 1, "req2": 2, "opt2": 2}
        assert exp_dict == {"req1": 1, "req2": 2}

        assert exp_dict != None
        assert exp_dict != {"req1": 1}
        assert exp_dict != {"req1": 1, "opt1": 1}
        assert exp_dict != {"req1": 1, "req2": 2, "opt3": 3}
        assert exp_dict != {"req1": 1, "req2": 2, "req3": 3}

        self.fail_validate_with({}, B.Dict({"req1": 1}),
                                "[e_msg=missing_required_field: req1] {} != Dict({'req1': 1})",
                                "[e_msg=missing_required_field: req1]")

        self.fail_validate_with({"req1": 1}, B.Dict({}),
                                "[e_msg=unknown_field] [e_unsafe_msg=unknown_field: req1] {'req1': 1} != Dict({})",
                                "[e_msg=unknown_field]",
                                )

        self.fail_validate_with({"req1": {}}, B.Dict({"req1": {"req11": 1}}),
                                "[e_path=$.req1] {} != {'req11': 1}",
                                "[e_path=$.req1]")

    def test_dict_matcher__DictOf(self):
        exp_dict = B.DictOf(B.AnyStr(), B.AnyStr())

        assert "DictOf(AnyStr(), AnyStr())" == str(exp_dict)

        assert exp_dict == {"1": "1"}
        assert exp_dict == {}
        assert exp_dict != {"1": 1}

        self.fail_validate_with({}, B.DictOf(B.AnyStr(), B.AnyStr(), n_item=1),
                                "[e_msg=mismatch_item_count: 0 != 1] {} != DictOf(AnyStr(), AnyStr(), n_item=1)",
                                "[e_msg=mismatch_item_count: 0 != 1]")

        self.fail_validate_with({}, B.DictOf(B.AnyStr(), B.AnyStr(), min_item=1),
                                "[e_msg=too_few_items: 0 < 1] {} != DictOf(AnyStr(), AnyStr(), min_item=1)",
                                "[e_msg=too_few_items: 0 < 1]")

        self.fail_validate_with({"1": "1", "2": "2"}, B.DictOf(B.AnyStr(), B.AnyStr(), max_item=1),
                                "[e_msg=too_many_items: 2 > 1] {'1': '1', '2': '2'} != DictOf(AnyStr(), AnyStr(), max_item=1)",
                                "[e_msg=too_many_items: 2 > 1]")

        self.fail_validate_with({"req1": 1}, B.DictOf(B.AnyStr(), B.AnyStr()),
                                "[e_path=$.req1] [e_msg=not_string] 1 != AnyStr()",
                                "[e_path=$.req1] [e_msg=not_string]",
                                )

        self.fail_validate_with({1: "req1"}, B.DictOf(B.AnyStr(), B.AnyStr()),
                                "[e_path=$.1] [e_msg=not_string] 1 != AnyStr()",
                                "[e_path=$.1] [e_msg=not_string]",
                                )

        self.fail_validate_with([], B.DictOf(B.AnyStr(), B.AnyStr()),
                                "[e_msg=not_dict] [] != DictOf(AnyStr(), AnyStr())",
                                "[e_msg=not_dict]")


class TestOtherMatcher(BaseClass):

    def test_other_matcher__Any(self):
        assert 'Any()' == str(B.Any())
        assert 'Any(str)' == str(B.Any(str))
        assert 'Any(str, bytes)' == str(B.Any(str, bytes))
        assert '1' == B.Any(str)
        assert 1 == B.Any(int)
        assert None != B.Any(str)
        assert '1' != B.Any(int)
        assert None == B.Any()

        self.fail_validate_with('1', B.Any(int),
                                "[e_msg=invalid_param] '1' != Any(int)",
                                "[e_msg=invalid_param]")

    def test_other_matcher__Not(self):
        assert 'Not(1)' == str(B.Not(1))
        assert 1 == B.Not(0)
        assert 0 != B.Not(0)

        self.fail_validate_with(0, B.Not(0),
                                "[e_msg=invalid_param] 0 != Not(0)",
                                "[e_msg=invalid_param]")

    def test_other_matcher__Nullable(self):
        assert 'Nullable(1)' == str(B.Nullable(1))
        assert 1 == B.Nullable(1)
        assert None == B.Nullable(1)
        assert 2 != B.Nullable(1)

        self.fail_validate_with(2, B.Nullable(1),
                                "[e_msg=invalid_param] 2 != Nullable(1)",
                                "[e_msg=invalid_param]")

    def test_BelieveMixin__complex(self):
        exp_dict = B.Dict({
            "company_info": B.Optional(
                B.ListOf(
                    B.Dict({
                        "attr": B.Dict({
                            "name": B.Any(str),
                            "id": B.Any(int)
                        })
                    })
                )
            )
        })

        self.fail_validate_with({"company_info": [{"attr": {"name": 123, "id": 123}}]}, exp_dict,
                                "[e_path=$.company_info.0.attr.name] [e_msg=invalid_param] 123 != Any(str)",
                                "[e_path=$.company_info.0.attr.name] [e_msg=invalid_param]")
        self.fail_validate_with({"company_info": [{"attr": {"name": "name"}}]}, exp_dict,
                                "[e_path=$.company_info.0.attr] [e_msg=missing_required_field: id] {'name': 'name'} != Dict({'name': Any(str), 'id': Any(int)})",
                                "[e_path=$.company_info.0.attr] [e_msg=missing_required_field: id]")



from typing import Any, List

from .internal import BelieveBase
from .internal import validate
from .internal import NO_CHECK


class OneOf(BelieveBase):
    def __init__(self, *candidates: Any) -> None:
        super().__init__(*candidates)
        self.__candidates = candidates

    def validate(self, rhs: Any, e_path: str = "") -> None:
        for i in self.__candidates:
            if i == rhs:
                return
        self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_argument")


class AnyOrder(BelieveBase):
    def __init__(self, list_obj: List) -> None:
        assert isinstance(list_obj, list)

        super().__init__(list_obj)
        self.__list_obj = list_obj

    def validate(self, rhs: List[Any], e_path: str = "") -> None:
        if not isinstance(rhs, list):
            self.raise_validate_error(rhs, e_path=e_path)
        if len(rhs) != len(self.__list_obj):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="different_length")
        list_obj = self.__list_obj[:]
        for idx, i in enumerate(rhs):
            if not i in list_obj:
                self.raise_validate_error(
                    rhs, e_path=e_path, e_msg=f"item_not_found_at_index: {idx}"
                )
            list_obj.remove(i)


class ListOf(BelieveBase):
    def __init__(
        self,
        one_item: Any,
        n_item: int = NO_CHECK,
        min_item: int = NO_CHECK,
        max_item: int = NO_CHECK,
    ) -> None:
        assert n_item == NO_CHECK or isinstance(n_item, int)
        assert min_item == NO_CHECK or isinstance(min_item, int)
        assert max_item == NO_CHECK or isinstance(max_item, int)

        super().__init__(one_item, n_item=n_item, min_item=min_item, max_item=max_item)

        self.__one_item = one_item
        self.__n_item = n_item
        self.__min_item = min_item
        self.__max_item = max_item

    def validate(self, rhs: List, e_path: str = "") -> None:
        if not isinstance(rhs, list):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="not_list")
        if self.__n_item != NO_CHECK:
            if not len(rhs) == self.__n_item:
                self.raise_validate_error(
                    rhs,
                    e_path=e_path,
                    e_msg=f"mismatch_item_count: {len(rhs)} != {self.__n_item}",
                )
        if self.__min_item != NO_CHECK:
            if len(rhs) < self.__min_item:
                self.raise_validate_error(
                    rhs,
                    e_path=e_path,
                    e_msg=f"too_few_items: {len(rhs)} < {self.__min_item}",
                )
        if self.__max_item != NO_CHECK:
            if len(rhs) > self.__max_item:
                self.raise_validate_error(
                    rhs,
                    e_path=e_path,
                    e_msg=f"too_many_items: {len(rhs)} > {self.__max_item}",
                )
        for idx, val in enumerate(rhs):
            validate(self.__one_item, val, "%s.%s" % (e_path, str(idx)))



import uuid
import json
from typing import Dict, List, Union
from urllib.parse import urlparse

from .internal import BelieveBase, NO_CHECK


class AnyStr(BelieveBase):
    def __init__(self, min_len: int = NO_CHECK, max_len: int = NO_CHECK) -> None:
        assert min_len == NO_CHECK or isinstance(min_len, int)
        assert max_len == NO_CHECK or isinstance(max_len, int)

        super().__init__(min_len=min_len, max_len=max_len)
        self.__min_len = min_len
        self.__max_len = max_len

    def validate(self, rhs: str, e_path: str = "") -> None:
        if not isinstance(rhs, str):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="not_string")
        if self.__min_len != NO_CHECK and len(rhs) < self.__min_len:
            self.raise_validate_error(
                rhs,
                e_path=e_path,
                e_msg=f"string_too_short: {len(rhs)} < {self.__min_len}",
            )
        if self.__max_len != NO_CHECK and len(rhs) > self.__max_len:
            self.raise_validate_error(
                rhs,
                e_path=e_path,
                e_msg=f"string_too_long: {len(rhs)} > {self.__max_len}",
            )


class AnyIntStr(BelieveBase):
    def validate(self, rhs: str, e_path: str = "") -> None:
        if not isinstance(rhs, str):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="not_string")

        try:
            int(rhs)
        except ValueError:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="not_int_string")


class AnyUUID(BelieveBase):
    def validate(self, rhs: str, e_path: str = "") -> None:
        if not isinstance(rhs, str) or len(rhs) != 36:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_uuid")
        try:
            uuid.UUID(rhs)
        except (TypeError, ValueError):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_uuid")


class AnyIPV4(BelieveBase):
    def validate(self, rhs: str, e_path: str = "") -> None:
        if not isinstance(rhs, str):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_ipv4")
        tokens = rhs.split(".")
        if len(tokens) != 4:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_ipv4")
        for token in tokens:
            try:
                token = int(token)
            except ValueError:
                self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_ipv4")
            if not 0 <= token < 256:
                self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_ipv4")


class AnySHA1(BelieveBase):
    def validate(self, rhs: str, e_path: str = "") -> None:
        if not isinstance(rhs, str):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_sha1")
        if len(rhs) != 40:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_sha1")
        try:
            int(rhs, 16)
        except ValueError:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_sha1")


class AnyJsonStr(BelieveBase):
    def __init__(self, json_obj: Union[Dict, List]) -> None:
        assert isinstance(json_obj, (dict, list))
        super().__init__(json_obj)
        self.__json_obj = json_obj

    def validate(self, rhs: str, e_path: str = "") -> None:
        if not isinstance(rhs, str):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="not_string")
        try:
            obj = json.loads(rhs)
        except:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="not_json_string")
        if obj != self.__json_obj:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="mismatch_json_string")


class AnyUrl(BelieveBase):
    def __init__(self, url: str) -> None:
        super().__init__(url)
        self.__parsed_url = urlparse(url)

    def _normalize_query(self, query: str) -> str:
        return "&".join(sorted(query.split("&")))

    def _default_port(self, scheme: str) -> int:
        if scheme == "http":
            return 80
        elif scheme == "https":
            return 443

    def validate(self, rhs: str, e_path: str = "") -> None:
        if not isinstance(rhs, str):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="not_string")

        o = urlparse(rhs)

        if self.__parsed_url.scheme != o.scheme:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="mismatch_scheme")

        default_port = self._default_port(self.__parsed_url.scheme)

        if self.__parsed_url.username != o.username:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="mismatch_username")

        if self.__parsed_url.password != o.password:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="mismatch_password")

        if self.__parsed_url.hostname != o.hostname:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="mismatch_hostname")

        if (self.__parsed_url.port or default_port) != (o.port or default_port):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="mismatch_port")

        if self.__parsed_url.path != o.path:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="mismatch_path")

        if self._normalize_query(self.__parsed_url.query) != self._normalize_query(
            o.query
        ):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="mismatch_query")



from typing import Union

from .internal import BelieveBase, NO_CHECK, USE_DEFAULT


class Almost(BelieveBase):
    def __init__(
        self, ts: Union[int, float], ts_range: Union[int, float] = USE_DEFAULT
    ) -> None:
        assert isinstance(ts, (int, float))
        assert ts_range == USE_DEFAULT or isinstance(ts_range, (int, float))
        super().__init__(ts, ts_range=ts_range)
        self.__ts = ts
        if ts_range == USE_DEFAULT:
            ts_range = 3
        self.__ts_range = ts_range

    def validate(self, rhs: Union[int, float], e_path: str = ""):
        if not isinstance(rhs, (int, float)):
            self.raise_validate_error(rhs, e_path=e_path, e_msg=f"not_int_or_float")
        if not self.__ts - self.__ts_range <= rhs <= self.__ts + self.__ts_range:
            self.raise_validate_error(rhs, e_path=e_path, e_msg=f"not_in_range")


class AnyInt(BelieveBase):
    def __init__(self, min_value: int = NO_CHECK, max_value: int = NO_CHECK) -> None:
        assert min_value == NO_CHECK or isinstance(min_value, int)
        assert max_value == NO_CHECK or isinstance(max_value, int)
        super().__init__(min_value=min_value, max_value=max_value)
        self.__min_value = min_value
        self.__max_value = max_value

    def validate(self, rhs: int, e_path: str = ""):
        if not isinstance(rhs, int):
            self.raise_validate_error(rhs, e_path=e_path)
        if self.__min_value != NO_CHECK and self.__min_value > rhs:
            self.raise_validate_error(
                rhs, e_path=e_path, e_msg=f"value_too_small: {rhs} < {self.__min_value}"
            )
        if self.__max_value != NO_CHECK and self.__max_value < rhs:
            self.raise_validate_error(
                rhs, e_path=e_path, e_msg=f"value_too_large: {rhs} > {self.__max_value}"
            )


class AnyFloat(BelieveBase):
    def __init__(
        self, min_value: float = NO_CHECK, max_value: float = NO_CHECK
    ) -> None:
        assert min_value == NO_CHECK or isinstance(min_value, float)
        assert max_value == NO_CHECK or isinstance(max_value, float)
        super().__init__(min_value=min_value, max_value=max_value)
        self.__min_value = min_value
        self.__max_value = max_value

    def validate(self, rhs: float, e_path: str = ""):
        if not isinstance(rhs, float):
            self.raise_validate_error(rhs, e_path=e_path)
        if self.__min_value != NO_CHECK and self.__min_value > rhs:
            self.raise_validate_error(
                rhs, e_path=e_path, e_msg=f"value_too_small: {rhs} < {self.__min_value}"
            )
        if self.__max_value != NO_CHECK and self.__max_value < rhs:
            self.raise_validate_error(
                rhs, e_path=e_path, e_msg=f"value_too_large: {rhs} > {self.__max_value}"
            )



import typing
from .internal import BelieveBase
from .internal import validate
from .internal import NO_CHECK


class Dict(BelieveBase):
    def __init__(self, dict_obj: typing.Dict) -> None:
        assert isinstance(dict_obj, dict)

        super().__init__(dict_obj)
        self.__dict_obj = dict_obj

    def validate(self, rhs: typing.Dict, e_path: str = "") -> None:
        if not isinstance(rhs, dict):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="not_dict")

        # validate required field
        for k, v in self.__dict_obj.items():
            if not isinstance(v, Optional):
                if k not in rhs:
                    self.raise_validate_error(
                        rhs, e_path=e_path, e_msg=f"missing_required_field: {k}"
                    )

        # validate field value
        for k, v in rhs.items():
            if k in self.__dict_obj:
                validate(self.__dict_obj[k], v, "%s.%s" % (e_path, k))
            else:
                self.raise_validate_error(
                    rhs,
                    e_path=e_path,
                    e_msg="unknown_field",
                    e_unsafe_msg=f"unknown_field: {k}",
                )


class DictOf(BelieveBase):
    def __init__(
        self,
        key: typing.Any,
        value: typing.Any,
        n_item: int = NO_CHECK,
        min_item: int = NO_CHECK,
        max_item: int = NO_CHECK,
    ) -> None:
        assert n_item == NO_CHECK or isinstance(n_item, int)
        assert min_item == NO_CHECK or isinstance(min_item, int)
        assert max_item == NO_CHECK or isinstance(max_item, int)

        super().__init__(
            key, value, n_item=n_item, min_item=min_item, max_item=max_item
        )
        self.__key = key
        self.__value = value
        self.__n_item = n_item
        self.__min_item = min_item
        self.__max_item = max_item

    def validate(self, rhs: Dict, e_path: str = "") -> None:
        if not isinstance(rhs, dict):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="not_dict")
        if self.__n_item != NO_CHECK:
            if not len(rhs) == self.__n_item:
                self.raise_validate_error(
                    rhs,
                    e_path=e_path,
                    e_msg=f"mismatch_item_count: {len(rhs)} != {self.__n_item}",
                )
        if self.__min_item != NO_CHECK:
            if len(rhs) < self.__min_item:
                self.raise_validate_error(
                    rhs,
                    e_path=e_path,
                    e_msg=f"too_few_items: {len(rhs)} < {self.__min_item}",
                )
        if self.__max_item != NO_CHECK:
            if len(rhs) > self.__max_item:
                self.raise_validate_error(
                    rhs,
                    e_path=e_path,
                    e_msg=f"too_many_items: {len(rhs)} > {self.__max_item}",
                )
        for k, v in rhs.items():
            validate(self.__key, k, "%s.%s" % (e_path, k))
            validate(self.__value, v, "%s.%s" % (e_path, k))


class Optional(BelieveBase):
    def __init__(self, value: typing.Any) -> None:
        super().__init__(value)
        self.__value = value

    def validate(self, rhs: typing.Any, e_path: str = "") -> None:
        validate(rhs, self.__value, e_path=e_path)



from typing import Any, List


class ValidateError(Exception):
    def __init__(self, v1: Any, v2: Any, **kwargs: Any) -> None:
        self.v1 = repr(v1)
        self.v2 = repr(v2)
        self.kwargs = kwargs

    def kwargs_to_string(self, fields: List[str], with_v2: bool) -> str:
        result = []
        for k in fields:
            if self.kwargs.get(k):
                if k == "e_path":
                    result.append(f"[{k}=${self.kwargs[k]}]")
                else:
                    result.append(f"[{k}={self.kwargs[k]}]")
        if with_v2:
            result.append(f"{self.v2} != {self.v1}")
        return " ".join(result)

    def xss_unsafe_message(self) -> str:
        return self.kwargs_to_string(["e_path", "e_msg", "e_unsafe_msg"], True)

    def xss_safe_message(self) -> str:
        return self.kwargs_to_string(["e_path", "e_msg"], False)

    def __str__(self) -> str:
        return self.xss_unsafe_message()



from .error import ValidateError

from .internal import BelieveBase
from .internal import validate

from .dict_matcher import Dict
from .dict_matcher import DictOf
from .dict_matcher import Optional

from .list_matcher import OneOf
from .list_matcher import AnyOrder
from .list_matcher import ListOf

from .number_matcher import Almost
from .number_matcher import AnyInt
from .number_matcher import AnyFloat

from .other_matcher import Nullable
from .other_matcher import Not
from .other_matcher import Any

from .str_matcher import AnyStr
from .str_matcher import AnyIntStr
from .str_matcher import AnyJsonStr
from .str_matcher import AnyUrl
from .str_matcher import AnyUUID
from .str_matcher import AnyIPV4
from .str_matcher import AnySHA1

# Put all Matcher into BelieveMixin
class BelieveMixin(object):
    pass


for c, cls in dict(locals()).items():
    try:
        if issubclass(cls, BelieveBase) and cls != BelieveBase:
            setattr(BelieveMixin, c, cls)
    except TypeError:
        pass

__version__ = "1.0.12"
